# import statements
#  probably want to import RN generator, CRC5 and CRC16
import G2_memoryBank
from time import *
import CRCs


class G2_LogicalTag:
	''' Instances of this class are logical tags.  A method is provided for each supported Gen2 command.  
	Each instance supports protocol tag attributes: SL, lastSession, currentQ, sessionFlags.   
	The tags also support a physical-logical attribute, CommandTime, which allows them to track
	flag persistence and reply timing.'''
	#
	def __init__(self):
		# set default values of basic attributes
		self.SL=False
		self.SLtarget=0	       #  default to any SL flag is ok
		self.lastSession='NULL'     # session is not meaningful until we receive a Query
		self.replyTime=[0,0]    #  reply time is [start,stop] in microseconds
		self.currentQ=0
		self.sessionFlags=['A','A','A','A']
		self.sessionTarget='A'		#  default target to A
		self.powerFlag=1
		self.lastCommand='NULL'	#  needed for Write
		self.lastReply=0		#  tagWorld time in microseconds when tag reply ended
		self.SLFlagtime=0		#  tagWorld time when SL flag is set
		self.sessionFlagtimes=[0,0,0,0]	#   tagWorld time when each session flag is set
		self.startTime=time()
		self.RTcal= 75		#  all time intervals in microseconds
		self.TRcal= 80
		self.Tari=25
		self.DivideRatio= 8
		self.Mvalue = 1	#  MMS multiplier
		self.Tpri=self.TRcal/self.DivideRatio	#  standard 1.0.9  table 6.3.1.3.3
		if self.RTcal>10*self.Tpri:
			self.T1=self.RTcal
		else:
			self.T1=10*self.Tpri
		self.T2max=20*self.Tpri
		self.TRext = 0
		self.LinkFreq = 1000/self.Tpri	# in KHz
		self.Truncate= 0    #  from Select command
		self.TruncatePointer=0

		#  persistence in microseconds, power on/off;  0 = lasts forever
		self.session0Persistence=[0,10]
		self.session1Persistence=[500000,500000]
		self.session2Persistence=[0,2000000]
		self.session3Persistence=[0,2000000]
		self.selectPersistence=[0,2000000]
		#      16-bit random numbers
		self.seed=self.startTime    #  initial value of the seed for random number generation
		self.QRN=0x00	#  this is the random number returned when the slot counter comes up 0
		self.handle=0x00	#  this is the random number returned by a Req_RN
		self.cover=0x00    	#  this is the current RN for cover coding
		self.parseDelay=10	#  time to parse a command in microseconds
		self.writeDelay=10000   #  time to complete a write to memory, in microseconds
		self.lockDelay=10000    #  time to complete set of lock bits, in microseconds
		self.slotCounter=0
		self.halfDead=0	#  this is used to note when the first of two
					#  KILL commands was successful 
		self.halfAccessed = 0	#  this is used to note when the first of two
					#  Access commands was successful
		self.state='off'		#  the tag needs to be powered up to do anything
		self.commandCount=0		#  keep track of total commands since inception
		
		#  make an instance of a tag memory bank
		self.tagMemory= G2_memoryBank.G2_memoryBank()	
		self.writeDefaultEPC()	
	
	#
	#	************************************************************
	#	SUPPORT FUNCTIONS
	#	************************************************************
	#
	def Checkmatch(self,Sel,Session,Target,CommandStopTime):
		'''This support function checks to see if the Select flag SL has the state requested by
		Sel (as defined in the standard's definition of e.g. the Query command), and verifies
		that for the given Session, the Target corresponds to the current state of the
		inventoried flag.  It returns True if everything matches and False if it doesn't.'''
		
		#  First check persistence status for power-on state...
		#
		if (self.selectPersistence[0]!=0):
			if (CommandStopTime-self.SLFlagtime>self.selectPersistence[0]):	# flag value lost, initialize
				self.SL=False
		
		if (self.session0Persistence[0]!=0):
			if(CommandStopTime-self.sessionFlagtimes[0])>self.session0Persistence[0]:   # flag lost, initialize
				self.sessionFlags[0]='A'
		
		if (self.session1Persistence[0]!=0):
			if (CommandStopTime-self.sessionFlagtimes[1])>self.session1Persistence[0]:   # flag lost, initialize
				self.sessionFlags[1]='A'

		if (self.session2Persistence[0]!=0):
			if (CommandStopTime-self.sessionFlagtimes[2])>self.session1Persistence[0]:   # flag lost, initialize
				self.sessionFlags[2]='A'

		if (self.session3Persistence[0]!=0):
			if (CommandStopTime-self.sessionFlagtimes[3])>self.session1Persistence[0]:   # flag lost, initialize
				self.sessionFlags[3]='A'
				
		
		
		
		
		
		#  first check to see if the SL flag state is correct
		SLmatch=False
		if ((Sel==0) or (Sel==1)):
			SLmatch=True
		if ((Sel==2) and (self.SL==False)):
			SLmatch=True
		if ((Sel==3) and (self.SL==True)):
			SLmatch=True
		
		#   then figure out if the target is correct
		if Target==0:
			targetflag='A'
		else:
			targetflag='B'
		if (targetflag==self.sessionFlags[Session]):
			targetmatch=True
		else:
			targetmatch=False
		if (SLmatch and targetmatch):
			return True
		else:
			return False
		
	def setLinkParameters(self,Tari,RTcal,TRcal,DR,M,TRext):
		''' This function sets the values of various link parameters.'''
		self.Tari=Tari
		self.TRcal=TRcal
		self.RTcal=RTcal
		self.Tpri=TRcal/DR
		self.LinkFreq=1000*DR/TRcal
		self.TRext=TRext
		self.Mvalue=M
		self.DivideRatio=DR
		if self.RTcal>10*self.Tpri:
			self.T1=self.RTcal
		else:
			self.T1=10*self.Tpri
		self.T2max=20*self.Tpri
	
	def setReplyTime(self,Time):
		'''This function sets the time at which the most recent reply occurred.'''
		self.lastReply=Time
		
	def checkReplyTime(self,Time):
		''' Check to see that the reply is within T2.'''
		
		difference=Time-self.lastReply
		print 'in checkReplyTime, time is:',Time,'  last reply was:',self.lastReply, 'difference:',difference
		if difference<self.T2max:
			return True
		else:
			return False
        
	def setSLFlag(self,SLvalue,Time):
		'''This function sets the select flag according to the value, and updates the time.'''
		self.SL=SLvalue
		self.SLFlagtime=Time
		
	def tagReply(self):
		'''This function assembles the reply to a valid ACK:  PC, EPC, CRC16, returned as a list of 16-bit words.
		Note truncate is not yet supported.'''
		
		PCbyte1=self.tagMemory.readMemoryByte(1,1,0)    #MSByte of PC word
		EPClength=int(PCbyte1/8)	#  pick out the first 5 bits; this is the length of the EPC in words
		PCbyte2=self.tagMemory.readMemoryByte(1,1,1)  #  LSByte of PC word
		PC=[256*PCbyte1+PCbyte2]
		EPC=[]
		for i in range(0,EPClength,1):  # accumulate the EPC
			EPC=EPC+[256*self.tagMemory.readMemoryByte(1,2+i,0)+self.tagMemory.readMemoryByte(1,2+i,1)]
		CRC=[256*self.tagMemory.readMemoryByte(1,0,0)+self.tagMemory.readMemoryByte(1,0,1)]
		return PC+EPC+CRC
	
	def getAccessPwd(self):
		'''This function returns the access password as an integer value.'''
		#   get bytes from tag memory;  recall we are MSword/MSbyte first
		byte1=self.tagMemory.readMemoryByte(0,2,0)
		byte2=	self.tagMemory.readMemoryByte(0,2,1)
		byte3=	self.tagMemory.readMemoryByte(0,3,0)
		byte4=	self.tagMemory.readMemoryByte(0,3,1)		
		
		#  now calculate the halves of the password
		pwd1=256*byte1+byte2
		pwd2=256*byte3+byte4
		return [pwd1,pwd2]	
		
	def getPC_EPC(self):
		'''This function reads the length of the EPC in the PC word and then returns a list
		with all the bytes of the PC and EPC.'''
		
		#  first get the high byte of the PC word
		PCbyte1=self.tagMemory.readMemoryByte(	1,1,0)
		#  then extract the value of the first 5 bits
		PCEPClength=int(PCbyte1/8)+1    #  see 6.3.2.1.4 of standard
		#
		returnlist=[]	# initialize an empty list
		for i in range(1,PCEPClength+1):
			highbyte=self.tagMemory.readMemoryByte(1,i,0)
			lowbyte=self.tagMemory.readMemoryByte(1,i,1)
			returnlist=returnlist+[highbyte,lowbyte]
		return returnlist
		
	def writeDefaultEPC(self):
		'''  This function writes PC and EPC, assuming a default GTIN carton
		with made-up manager, and RN for serial number.  Partition 5 = 24-bit company.'''
		PCbyte1=0x30	#  96-bit EPC
		PCbyte2=0x00
		Header=48	#  GTIN map
		Type=3	#  case
		Partition = 5	#  24-bit company name
		Companybyte1=0x00		#randomly chosen manager (company) number
		Companybyte2=0x10
		Companybyte3=0x51
		MNhighbyte=0x05		# object class (model number)
		MNlowbyte=0x31
		MNnibble=0xe
		SNhighbits=0x0		#  first two bits
		SNhighnibble=0x5		#  next 4 bits
		SNfirstRN=CRCs.RN16(int(100*self.startTime))
		SNsecondRN=CRCs.RN16(SNfirstRN)
		
		return1=self.tagMemory.writeMemoryByte(1,1,0,PCbyte1)
		return2=self.tagMemory.writeMemoryByte(1,1,1,PCbyte2)
		return3=self.tagMemory.writeMemoryByte(1,2,0,Header)
		
		#  now we have to partition the bits into bytes
		
		byte4=Type*32+Partition*4+int(Companybyte1/64)
		byte5=(Companybyte1%64)*4+int(Companybyte2/64)
		byte6=(Companybyte2%64)*4+int(Companybyte3/64)
		byte7=(Companybyte3%64)*4+int(MNhighbyte/64)
		byte8=(MNhighbyte%64)*4+int(MNlowbyte/64)
		byte9=(MNlowbyte%64)*4+int(MNnibble/4)
		byte10=(MNnibble%4)*64+SNhighbits*16+SNhighnibble
		byte11=int(SNfirstRN/256)
		byte12=SNfirstRN%256
		byte13=int(SNsecondRN/256)
		byte14=SNsecondRN/256
		
		return4=self.tagMemory.writeMemoryByte(1,2,1,byte4)
		return5=self.tagMemory.writeMemoryByte(1,3,0,byte5)
		return6=self.tagMemory.writeMemoryByte(1,3,1,byte6)
		return7=self.tagMemory.writeMemoryByte(1,4,0,byte7)
		return8=self.tagMemory.writeMemoryByte(1,4,1,byte8)
		return9=self.tagMemory.writeMemoryByte(1,5,0,byte9)
		return10=self.tagMemory.writeMemoryByte(1,5,1,byte10)
		return11=self.tagMemory.writeMemoryByte(1,6,0,byte11)
		return12=self.tagMemory.writeMemoryByte(1,6,1,byte12)
		return13=self.tagMemory.writeMemoryByte(1,7,0,byte13)
		return14=self.tagMemory.writeMemoryByte(1,7,1,byte14)

	def replyDuration(self,dataBits):
		'''This function computes the duration of a reply given
		the current values of M,Tpri,TRext.'''

		#  first compute the bit duration
		bitTime=self.Tpri*self.Mvalue
			#  the duration of one bit in microseconds

		#  now find the duration of the preamble
		if self.Mvalue==0:      #  FM0
			if self.TRext==0:
				preambleTime=6*bitTime          # no pilot tone
			else:
				preambleTime=18*bitTime         # pilot tone
		else:                   #  MMS
			if self.TRext==0:
				preambleTime=6*bitTime          # no pilot tone
			else:
				preambleTime=10*bitTime         # pilot tone

		#  and finally add the time to send the data

		return (preambleTime+dataBits*bitTime)
		
		
	#
	#	************************************************************
	#	COMMAND FUNCTIONS
	#	************************************************************
	#
	
		
	#
	#	************************************************************
	#	QueryRep
	#	************************************************************
	#
	
	def QueryRep(self,Session,CommandTime):
		'''  Implement the QueryRep command, v.1.0.9 and appendix C.   CommandTime
		is the tagWorld time in microseconds.'''

		passiveReplyTime=[CommandTime[1],CommandTime[1]+self.parseDelay]
		offReplyTime=[CommandTime[1],CommandTime[1]]
		replyTime=passiveReplyTime   #  default
		
		
		if self.state!='off':	#  the tag does not respond at all when it is off
		
			self.commandCount=self.commandCount+1		# increment command count since creation
			self.lastCommand='QueryRep'
			
			if ((self.state=='reply') or (self.state=='acknowledged')):
				if self.checkReplyTime(CommandTime[0])==False:		# if time since last reply > T2
					self.state='arbitrate'
					return ['NULL',passiveReplyTime]	

	
			#   for Session=lastSession ...
			#   if ready just stay in ready
			#   if arbitrate and slot-1 is not 0, stay in arbitrate
			#   if arbitrate and slot-1 is 0, backscatter new RN16, go to reply
			#   if reply, go to arbitrate
			#   if acknowledged, open, secured, flip inventoried flag, go to ready
			#   if killed do nothing
			
			#  for Session not equal to lastSession
			#  if ready, arbitrate, reply, aknowledged:  do nothing
			#  if open stay in open UNLESS halfdead or halfaccessed in which case go to arbitrate
			#  if secured stay in secured UNLESS halfdead or halfaccessed in which case go to arbitrate
			#  if killed do nothing
			if (Session==self.lastSession):	# session is the same as last session
				if (self.state=='ready'):
					
					return ['NULL',passiveReplyTime]
	
						#  the second entry of the tuple is the response delay in microseconds
				if (self.state=='arbitrate'):
					#  decrement the slot counter
					self.slotCounter=self.slotCounter-1
					if self.slotCounter<0:
						self.slotCounter=0x7FFF	#  see page 43 of 1.0.9

						return ['NULL',passiveReplyTime]
					#   return a new RN16 if slot is 0 
					if (self.slotCounter==0):		#  return an RN and go to 'reply'
						self.state='reply'
						self.QRN=CRCs.RN16(self.seed)	# get a 16-bit random number
						self.seed=CRCs.RN16(self.handle)  # hash it to produce the next seed
						replyTime[0]=CommandTime[1]+self.T1
						replyTime[1]=replyTime[0]+self.replyDuration(16)

							# store time of most recent reply for T2 timing
						self.setReplyTime(replyTime[1])
						return [self.QRN,replyTime]		
							#  return the 16-bit RN with delay
		
						#   if slot is not 0 do nothing
					else:
						return ['NULL',passiveReplyTime]

				if (self.state=='reply'):		#  no need to check time here since we go to arbitrate anyway
					self.state='arbitrate'
					return ['NULL',passiveReplyTime]	
				if ((self.state=='open') or (self.state=='secured')):
					#    flip the inventoried flag for this session
					if self.sessionFlags[Session]=='A':    
						self.sessionFlags[Session]='B'
					else:
						self.sessionFlags[Session]='A'
					self.sessionFlagtimes[Session]=CommandTime
					self.state='ready'
					return ['NULL',passiveReplyTime]
					
				if (self.state=='acknowledged'):
					#    flip the inventoried flag for this session
					if self.sessionFlags[Session]=='A':    
						self.sessionFlags[Session]='B'
					else:
						self.sessionFlags[Session]='A'
					self.sessionFlagtimes[Session]=CommandTime
					self.state='ready'
					return ['NULL',passiveReplyTime]						
				if (self.state=='killed'):
					return ['NULL',offReplyTime]
					
				
			else:	#   case of this session not the same as the last session...
				if  ((self.state=='open') or (self.state=='secured')):
					if ((self.halfDead==1) or (self.halfAccessed==1)):	
							# we were in the middle of a Kill or Access
						self.state='arbitrate'				
							#  so abort whatever it was and return to arbitrate
						self.halfDead=0
						self.halfAccessed=0
				return ['NULL',passiveReplyTime]

			
		#end case self.state!='off'
		else:

			return ['NULL',offReplyTime]
	#
	#	************************************************************
	#	ACK
	#	************************************************************
	#
	
	def ACK(self,RN16,CommandTime):
		''' Respond to an ACK from the reader.'''

		passiveReplyTime=[CommandTime[1],CommandTime[1]+self.parseDelay]
		offReplyTime=[CommandTime[1],CommandTime[1]]
		replyTime=passiveReplyTime
		
		if self.state!='off':	#  the tag does not respond at all when it is off
			self.commandCount=self.commandCount+1		# increment command count since creation
			self.lastCommand='ACK'
			
			if ((self.state=='reply') or (self.state=='acknowledged')):
				if self.checkReplyTime(CommandTime[0])==False:		# if time since last reply > T2
					self.state='arbitrate'
				else:
					if (RN16==self.QRN):
						replyByteList=self.tagReply()
						replyTime[0]=CommandTime[1]+self.T1
						replyTime[1]=replyTime[0]+self.replyDuration(8*len(replyByteList))

						self.setReplyTime(replyTime[1])
						self.state='acknowledged'
						return [replyByteList,replyTime]
					else:
						self.state='arbitrate'
						return ['NULL',passiveReplyTime]
			
			if ((self.state=='ready') or (self.state=='arbitrate')):
				return ['NULL',passiveReplyTime]
				
			if ((self.state=='open') or (self.state=='secured')):
				if (RN16==self.handle):
					replyByteList=self.tagReply()
					replyTime[0]=CommandTime[1]+self.T1
					replyTime[1]=replyTime[0]+self.replyDuration(8*len(replyByteList))
					self.setReplyTime(replyTime[1])
					return [replyByteList,replyTime]   #  reply but don't change state
				else:
					self.state='arbitrate'
					return ['NULL',passiveReplyTime]
		
			if (self.state=='killed'):
				return ['NULL',offReplyTime]	#  do nothing
				
				
		# end case of  self.state!='off'
		else:
			return  ['NULL',offReplyTime]	
		
	#
	#	************************************************************
	#	Query
	#	************************************************************
	#	
		
	def Query(self,Tari,RTcal,TRcal,DR,M,TRext,Sel,Session,Target,Q,CRC5,CommandTime):
		'''This method implements the Query command, based on v 1.0.9 p. 49 ff and appendix C command tables.
		 It returns an RN16 if appropriate,   otherwise it returns 'NULL', here equivalent to the standard's 
		requirement for 'no reply'.  '''
		

		passiveReplyTime=[CommandTime[1],CommandTime[1]+self.parseDelay]
		offReplyTime=[CommandTime[1],CommandTime[1]]
		replyTime=passiveReplyTime

		if self.state!='off':	#  the tag does not respond at all when it is off

			self.commandCount=self.commandCount+1		# increment command count since creation
			self.lastCommand='Query'
			
			if ((self.state=='reply') or (self.state=='acknowledged')):
				if self.checkReplyTime(CommandTime[0])==False:		# if time since last reply > T2
					self.state='arbitrate'
					
					
			if ((self.state=='ready') or (self.state=='arbitrate') or (self.state=='reply')):
				self.setLinkParameters(Tari,RTcal,TRcal,DR,M,TRext)
				
				self.SLtarget=Sel		#  save the targets for this Query
				self.sessionTarget=Target
				
				#   we need to initialize the slot counter first...
				self.currentQ=Q
				self.slotCounter=CRCs.RNSlot(Q,self.seed)
				self.seed=CRCs.RN16(self.seed)	#  each time we use the seed we hash it to produce the new seed
				
				#   return a new RN16 if slot is 0 and inventory / select flags match and go to reply
				if self.Checkmatch(Sel,Session,Target,CommandTime[1]):		#   SL and inventoried flags match
					if (self.slotCounter==0):		#  return an RN and go to 'reply'
						self.state='reply'
						self.QRN=CRCs.RN16(self.seed)	# get a 16-bit random number
						self.seed=CRCs.RN16(self.handle)  # hash it to produce the next seed
						self.lastSession=Session	#  this is now the current session

						replyTime[0]=CommandTime[1]+self.T1
						replyTime[1]=replyTime[0]+self.replyDuration(16)

							# store time of most recent reply for T2 timing
						self.setReplyTime(replyTime[1])

						#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
						#print 'Command time is:', CommandTime,' and T1 is:', self.T1
                                                        
						return [self.QRN,replyTime]     #  return the 16-bit RN with delay


					#   if slot is not 0 but targets match return 'NULL', go to arbitrate
					else:

						self.state='arbitrate'
						self.lastSession=Session	#this is now the current session
						return  ['NULL',passiveReplyTime]
				
				#   if targets don't match return 'NULL', go to ready
				else:
					self.state='ready'
					self.lastSession=Session	#   does the last session update even when you don't match???
					return  ['NULL',passiveReplyTime]
				
				
			if ((self.state=='acknowledged') or (self.state=='open') or (self.state=='secured')):
				#!!!!!!!!!!!!!!   debug
				# print('in query the last session was:',self.lastSession,' and the current session is:',Session)
				
				if Session==self.lastSession:			#  same as most recent session, so flip flag for that session
					# print('in query the last flag was',self.sessionFlags[Session])
					if self.sessionFlags[Session]=='A':    
						self.sessionFlags[Session]='B'
					else:
						self.sessionFlags[Session]='A'
					
				#   we need to initialize the slot counter first...
				self.slotCounter=CRCs.RNSlot(Q,self.seed)
				self.currentQ=Q
				self.seed=CRCs.RN16(self.seed)	#  each time we use the seed we hash it to produce the new seed
				
				#   return a new RN16 if slot is 0 and inventory / select flags match and go to reply
				if self.Checkmatch(Sel,Session,Target,CommandTime[1]):		#   SL and inventoried flags match
					if (self.slotCounter==0):		#  return an RN and go to 'reply'
						self.state='reply'
						self.QRN=CRCs.RN16(self.seed)	# get a 16-bit random number
						self.seed=CRCs.RN16(self.handle)  # hash it to produce the next seed
						self.lastSession=Session	#  this is now the current session

						replyTime[0]=CommandTime[1]+self.T1
						replyTime[1]=replyTime[0]+self.replyDuration(16)

							# store time of most recent reply for T2 timing
						self.setReplyTime(replyTime[1])
						return [self.QRN,replyTime]		
							#  return the 16-bit RN with delay T1


					#   if slot is not 0 but targets match return 'NULL', go to arbitrate
					else:

						self.state='arbitrate'
						self.lastSession=Session	#this is now the current session
						return  ['NULL',passiveReplyTime]
				
				#   if targets don't match return 'NULL', go to ready
				else:
					self.state='ready'
					self.lastSession=Session	#   does the last session update even when you don't match???
					return  ['NULL',passiveReplyTime]
				
			if ((self.state=='killed')):  return  ['NULL',offReplyTime]
			
		#  end case of self.state!='off'
		else:
			return  ['NULL',offReplyTime]
	
	#  end Query
	

		
	#
	#	************************************************************
	#	QueryAdjust
	#	************************************************************
	#	
		
	def QueryAdjust(self,Session,UpDn,CommandTime):
		'''Implement the QueryAdjust command, based on version 1.0.9 page 50 ff. and appendix C.  '''


		passiveReplyTime=[CommandTime[1],CommandTime[1]+self.parseDelay]
		offReplyTime=[CommandTime[1],CommandTime[1]]
		replyTime=passiveReplyTime

		
		if self.state!='off':	#  the tag does not respond at all when it is off
		
			self.commandCount=self.commandCount+1		# increment command count since creation
			self.lastCommand='QueryAdjust'
			
			if ((self.state=='reply') or (self.state=='acknowledged')):
				if self.checkReplyTime(CommandTime[0])==False:		# if time since last reply < T2
					self.state='arbitrate'
					
			#   if the session is different from that of the last Query we ignore the command
			#
			#   if an invalid UpDn value is received, ignore the command
			validFlag=False
			if (UpDn==6):
				Q=self.currentQ+1
				if Q==16: Q=15	# don't exceed limit
				validFlag=True
			if (UpDn==0):
				Q=self.currentQ
				validFlag=True
			if (UpDn==3):
				Q=self.currentQ-1
				if Q<0:  Q=0		#  don't exceed limit
				validFlag=True
	
			if ((Session==self.lastSession) and validFlag):	
					#  a valid UpDn value was received and session is the same as last session
				if (self.state=='ready'):
					return ['NULL',CommandTime[1]+self.parseDelay]
				if ((self.state=='arbitrate') or (self.state=='reply')):
					#   we need to initialize the slot counter first...
					self.currentQ=Q
					self.slotCounter=CRCs.RNSlot(Q,self.seed)
					self.seed=CRCs.RN16(self.seed)	#  each time we use the seed we hash it to produce the new seed
					
					#   return a new RN16 if slot is 0 and inventory / select flags match and go to reply
					if self.Checkmatch(self.SLtarget,Session,self.sessionTarget,CommandTime[1]):	
							#   SL and inventoried flags match
						if (self.slotCounter==0):		#  return an RN and go to 'reply'
							self.state='reply'
							self.QRN=CRCs.RN16(self.seed)	# get a 16-bit random number
							self.seed=CRCs.RN16(self.handle)  # hash it to produce the next seed


							replyTime[0]=CommandTime[1]+self.T1
							replyTime[1]=replyTime[0]+self.replyDuration(16)

								# store time of most recent reply for T2 timing
							self.setReplyTime(replyTime[1])

							return [self.QRN, replyTime]	#  return the 16-bit RN
	
							#   if slot is not 0 but targets match return 'NULL', go to arbitrate
						else:
							self.state='arbitrate'
							return ['NULL',passiveReplyTime]
					
					#   if targets don't match return 'NULL', go to ready
					else:
						self.state='ready'
						return ['NULL',CpassiveReplyTime]
							
				if ((self.state=='acknowledged') or (self.state=='open') or (self.state=='secured')):
					#    flip the inventoried flag for this session
					if self.sessionFlags[Session]=='A':    
						self.sessionFlags[Session]='B'
					else:
						self.sessionFlags[Session]='A'
					self.sessionFlagtimes[Session]=CommandTime[1]
					self.state='ready'
					return ['NULL',passiveReplyTime]
						
				if ((self.state=='killed')):  return ['NULL',offReplyTime]
					
				
			else:	#   case of this session not the same as the last session...
				if  ((self.state=='open') or (self.state=='secured')):
					if ((self.halfDead==1) or (self.halfAccessed==1)):	# we were in the middle of a Kill or Access
						self.state='arbitrate'					#  so abort whatever it was 
						self.halfDead=0					#  and return to arbitrate
						self.halfAccessed=0
				return ['NULL',passiveReplyTime]
				if ((self.state=='killed')):  return ['NULL',offReplyTime]
				
		# end case of self.state!='off'
		else:
			return  ['NULL',offReplyTime]
	
		
	#
	#	************************************************************
	#	Select
	#	************************************************************
	#		
		
	def Select(self,Target,Action,Membank,Pointer,Length,Mask,Truncate,CRC16,CommandTime):
		'''implement the Select command, based on version 1.0.9 page 47ff.  
		Target is an integer in this layer NOT a bitstream.  Mask is a list of bits.
		Note EBV is NOT supported in this layer; Pointer is just a number.'''
		# 
		# 
		#

		passiveReplyTime=[CommandTime[1],CommandTime[1]+self.parseDelay]
		offReplyTime=[CommandTime[1],CommandTime[1]]
		replyTime=passiveReplyTime

		if self.state!='off':	#  the tag does not respond at all when it is off
		
		
			self.commandCount=self.commandCount+1		# increment command count since creation
			self.lastCommand='Select'
			
			if ((self.state=='reply') or (self.state=='acknowledged')):
				if self.checkReplyTime(CommandTime[0])==False:		# if time since last reply > T2
					self.state='arbitrate'
				else:
					self.state='ready'	#  note flag does not flip
					
			if ((self.state=='ready') or (self.state=='arbitrate') or (self.state=='open') or (self.state=='secured')):
				self.state='ready'
			
			print 'in Select, Memory bank:',Membank,' Pointer:', Pointer, 'Length:', Length, ' Mask:', Mask
			ignore=0
			if ((Membank<1) or (Membank>4) or (Target>4)):    #  can't match to pwds or no bank; Target>4 RFU
				ignore=1
			if ((Truncate==1) and (Target!=4)):		#  truncate only applies if target is Select
				ignore=1
			if ((Truncate==1) and (Membank!=1)):	#   truncate only applies if in EPC memory
				ignore=1
			if ignore==0:
				#  process the command, else do nothing
				match=0	#   default is no match
				if Pointer>=0:		# negative pointer can't match
					if ((Pointer+Length)<256):    #  the mask is in memory
						if Length==0:
							match=1	#  zero length means anything matches
						else:
							try:
								if self.tagMemory.matchMemory(Membank,Pointer,Length,Mask):
									match=1
							except ValueError:
								match=0	
									#  if we generate an error due to mismatched mask length, can't match
					if (Truncate==0):		#  Truncate data to default
						self.Truncate=0
						self.TruncatePointer=0
					if (Truncate==1):
						if ((Pointer+Length)>(32+self.EPClength)):
							match=0	
						else:   #  enable truncate for next ACK
							self.Truncate=1
							self.TruncatePointer=Pointer+Length
				if match==1:
					print 'mask matched, target is:', Target, ' and action is:', Action
					if ((Target>-1) and (Target<4)):    #  one of the session flags
						if Action==0: 
							self.sessionFlags[Target]='A'
							self.sessionFlagtimes[Target]=CommandTime[1]+self.parseDelay
						if Action==1: 
							self.sessionFlags[Target]='A'
							self.sessionFlagtimes[Target]=CommandTime[1]+self.parseDelay
						if Action==2:  #  do nothing
							pass
						if Action==3: 
							if self.sessionFlags[Target]=='A':
								self.sessionFlags[Target]='B'
							else:
								self.sessionFlags[Target]='A'
							self.sessionFlagtimes[Target]=CommandTime[1]+self.parseDelay
						if Action==4: 
							self.sessionFlags[Target]='B'
							self.sessionFlagtimes[Target]=CommandTime[1]+self.parseDelay
						if Action==5: 
							self.sessionFlags[Target]='B'
							self.sessionFlagtimes[Target]=CommandTime[1]+self.parseDelay
						if Action==6:  #   do nothing
							pass
						if Action==7:  #   do nothing
							pass
					if Target==4:	#  the Select flag
						if Action==0:  
							self.SL=True
							self.SLFlagtime=CommandTime[1]+self.parseDelay
						if Action==1:  
							self.SL=True
							self.SLFlagtime=CommandTime[1]+self.parseDelay
						if Action==2:  #   do nothing
							pass
						if Action==3:   
							if self.SL==True:
								self.SL=False
							else:
								self.SL=True
							self.SLFlagtime=CommandTime[1]+self.parseDelay
						if Action==4:	
							self.SL=False
							self.SLFlagtime=CommandTime[1]+self.parseDelay
						if Action==5:   
							self.SL=False
							self.SLFlagtime=CommandTime[1]+self.parseDelay
						if Action==6:  #   do nothing
							pass
						if Action==7:  #   do nothing
							pass
				if match==0:
					if ((Target>-1) and (Target<4)):    #  one of the session flags
						if Action==0: 
							self.sessionFlags[Target]='B'
							self.sessionFlagtimes[Target]=CommandTime[1]+self.parseDelay
						if Action==1: #   do nothing
							pass
						if Action==2:  
							self.sessionFlags[Target]='B'
							self.sessionFlagtimes[Target]=CommandTime[1]+self.parseDelay
						if Action==3: #   do nothing
							pass
						if Action==4: 
							self.sessionFlags[Target]='A'
							self.sessionFlagtimes[Target]=CommandTime[1]+self.parseDelay
						if Action==5: #   do nothing
							pass
						if Action==6: 
							self.sessionFlags[Target]='A'
							self.sessionFlagtimes[Target]=CommandTime[1]+self.parseDelay
						if Action==7: 
							if self.sessionFlags[Target]=='A':
								self.sessionFlags[Target]='B'
							else:
								self.sessionFlags[Target]='A'
							self.sessionFlagtimes[Target]=CommandTime[1]+self.parseDelay
							
					if Target==4:	#  the Select flag
						if Action==0:  
							self.SL=False
							self.SLFlagtime=CommandTime[1]+self.parseDelay
						if Action==1:  #   do nothing
							pass
						if Action==2:  
							self.SL=False
							self.SLFlagtime=CommandTime[1]+self.parseDelay
						if Action==3:  #  do nothing
							pass
						if Action==4:   
							self.SL=True
							self.SLFlagtime=CommandTime[1]+self.parseDelay
						if Action==5:  #  do nothing
							pass
						if Action==6:   
							self.SL=True
							self.SLFlagtime=CommandTime[1]+self.parseDelay
						if Action==7: 
							if self.SL==True:
								self.SL=False
							else:
								self.SL=True	
							self.SLFlagtime=CommandTime[1]+self.parseDelay
							
			return ['NULL',passiveReplyTime]		#  equivalent of no reply in standard
								
		#  end case of self.state!='off'	
		else:
			return  ['NULL',offReplyTime]
		
	#
	#	************************************************************
	#	NAK
	#	************************************************************
	#	
			
	def NAK(self,CommandTime):
		'''  Implement NAK as per 1.0.9 ...   All states except ready and killed are sent to arbitrate.'''


		passiveReplyTime=[CommandTime[1],CommandTime[1]+self.parseDelay]
		offReplyTime=[CommandTime[1],CommandTime[1]]
		replyTime=passiveReplyTime
		
		if self.state!='off':
			
			self.commandCount=self.commandCount+1		# increment command count since creation
			self.lastCommand='NAK'
			
			if ((self.state=='arbitrate') or (self.state=='reply') or (self.state=='acknowledged') or (self.state=='open') or (self.state=='secured')):
				self.state='arbitrate'
				return ['NULL',passiveReplyTime]
			
			if self.state=='ready':
				return ['NULL',passiveReplyTime]	#  do nothing, leave us in ready
				
			if self.state=='killed':
				return ['NULL',offReplyTime]	
				
		# end case of self.state!='off':		
		else:
			return  ['NULL',offReplyTime]
		
		
	#
	#	************************************************************
	#	Req_RN
	#	************************************************************
	#
	
	def Req_RN(self,RN16,CRC16,CommandTime):
		'''   Implement Req_RN as per 1.0.9.   '''


		passiveReplyTime=[CommandTime[1],CommandTime[1]+self.parseDelay]
		offReplyTime=[CommandTime[1],CommandTime[1]]
		replyTime=passiveReplyTime
		
		if self.state!='off':
			
			self.commandCount=self.commandCount+1		# increment command count since creation
			self.lastCommand='Req_RN'
			
			if ((self.state=='reply') or (self.state=='acknowledged')):
				if self.checkReplyTime(CommandTime[0])==False:		# if time since last reply < T2
					self.state='arbitrate'
			
			if self.state=='ready':
				return ['NULL',passiveReplyTime]	#  do nothing, stay in ready
			
			if ((self.state=='arbitrate') or (self.state=='reply')):
				self.state='arbitrate'
				return ['NULL',passiveReplyTime]
			
			if self.state=='acknowledged':
				if RN16==self.QRN:		#  the RN16 provided is the same as that we backscattered
									#  in response to Query/Rep/Adj
					pwdlist=self.getAccessPwd()
					if ((pwdlist[0]==0) and (pwdlist[1]==0)):	#  0-value password
						self.state='secured'
					else:
						self.state='open'	#  non-zero password
				
					self.handle=CRCs.RN16(self.seed)	# get a 16-bit random number

					replyTime[0]=CommandTime[1]+self.T1
					replyTime[1]=replyTime[0]+self.replyDuration(32)   # note physical reply includes CRC!
											#  even though we don't send it here
						# store time of most recent reply for T2 timing
					self.setReplyTime(replyTime[1])
					self.seed=CRCs.RN16(self.handle)  # hash it to produce the next seed
					return [self.handle, replyTime]	#  return the 16-bit RN
	
				else:		#  case of RN16 not the same as the most recent RN16
					return ['NULL',CommandTime[1]+self.parseDelay]	# do nothing
				
			if ((self.state=='open') or (self.state=='secured')):
				if RN16==self.handle:		#  the RN16 provided is the same as that we backscattered
									#  in response to Query/Rep/Adj
					self.cover=CRCs.RN16(self.seed)	#  get a 16-bit random number
					self.seed=CRCs.RN16(self.cover)	#  hash it to produce the next seed

					replyTime[0]=CommandTime[1]+self.T1
					replyTime[1]=replyTime[0]+self.replyDuration(32)	# note physical reply
											# includes CRC! even though we don't send here
						# store time of most recent reply for T2 timing
					self.setReplyTime(replyTime[1])

					return [self.cover,replyTime]		#  return the cover word
	
				else:
					return ['NULL',passiveReplyTime]	#  do nothing
					
			if self.state=='killed':
				return ['NULL',offReplyTime]	# do nothing
			
		# end case of self.state!='off':		
		else:
			return  ['NULL',offReplyTime]
		
	#
	#	************************************************************
	#	Read
	#	************************************************************
	#
	
	def Read(self,Membank,WordPtr,WordCount,RN16,CRC16,CommandTime):
		'''  Implement read command as per standard 6.3.2.10.3.2 p. 56, tables in appendix C.'''

		passiveReplyTime=[CommandTime[1],CommandTime[1]+self.parseDelay]
		offReplyTime=[CommandTime[1],CommandTime[1]]
		replyTime=passiveReplyTime
		
		if self.state!='off':
			
			self.commandCount=self.commandCount+1		# increment command count since creation
			self.lastCommand='Read'
			
			if self.state=='ready':
				return ('NULL',passiveReplyTime)		# do nothing
			#  for the moment we just reuse provided CRC16 -- need to implement check and regeneration!
			
			if ((self.state=='reply') or (self.state=='acknowledged') or (self.state=='arbitrate')): 
				self.state='arbitrate'
				return ('NULL',passiveReplyTime)
					
			if ((self.state=='open') or (self.state=='secured')):
				#   read the data unless the info is locked or the location is outside the memory bank
				if (RN16==self.handle):	#  process the Read command
					try:	#  attempt the read, return an error if the relevant info is locked
						#  or the requested words are outside memory bounds
						
						returnlist=[]		# initialize an empty list
						locked=0		# initialize locked state
						if Membank==0:
							if (WordPtr<2):	#  this is a request for the kill password
								if self.tagMemory.lockKill==1:
									locked=1
							if ((WordPtr<4) or (WordPtr+WordCount<4)):   #  this request includes the access password
								if self.tagMemory.lockAccess==1:
									locked=1
						if WordCount==0:	#  backscatter the whole memory bank or EPC
							if Membank==1:	#  if WordPtr in EPC return EPC, else return all bank
										#  after WordPtr
								#
								if ((WordPtr>0) and (WordPtr<PCEPClength)):  # ptr is in EPC
									returnlist=self.getPC_EPC()
								else:   #  ptr not in EPC
									for i in range(WordPtr,self.tagMemory.banksize):
										highbyte=self.tagMemory.readMemoryByte(Membank,i,0)
										lowbyte=self.tagMemory.readMemoryByte(Membank,i,1)
										returnlist=returnlist+[256*highbyte+lowbyte]
								
							else:			# WordPtr is not in the EPC so backscatter all bank after WordPtr
								for i in range(WordPtr,self.tagMemory.banksize):
									if locked==0:	#  backscatter a reply
										highbyte=self.tagMemory.readMemoryByte(Membank,i,0)
										lowbyte=self.tagMemory.readMemoryByte(Membank,i,1)
										returnlist=returnlist+[256*highbyte+lowbyte]	
						else:	#  WordCount is not 0 
							for i in range(WordPtr,WordPtr+WordCount):
								if locked==0:   #  backscatter a reply
									highbyte=self.tagMemory.readMemoryByte(Membank,i,0)
									lowbyte=self.tagMemory.readMemoryByte(Membank,i,1)
									returnlist=returnlist+[256*highbyte+lowbyte]	
						if locked==0:
							replyTime[0]=CommandTime[1]+self.T1
							replyTime[1]=replyTime[0]+self.replyDuration(1+len(returnlist)*16+32)
							enclosingList=[0,returnlist,RN16,CRC16,replyTime]
						else:
							replyTime[0]=CommandTime[1]+self.T1
							replyTime[1]=replyTime[0]+self.replyDuration(1+1+32)
							enclosingList=[1,4,RN16,CRC16,replyTime]	
								#  return error code for locked memory
					except ValueError:
						replyTime[0]=CommandTime[1]+self.T1
						replyTime[1]=replyTime[0]+self.replyDuration(1+1+32)
						enclosingList=[1,3,RN16,CRC16,replyTime] 
							#  return error code for memory overrun
					#print 'in Read returned:', enclosingList
					return enclosingList
				else:
					# ignore the command if invalid handle, stay in same state
					return ('NULL',passiveReplyTime)	
			
			if self.state=='killed':
				return ('NULL',offReplyTime)
		
		#end case of self.state!='off':
		else:
			return  ['NULL',offReplyTime]
	#
	#	************************************************************
	#	Write
	#	************************************************************
	#
	
	def Write(self,Membank,WordPtr,Data,RN16,CRC16,CommandTime):
		''' Implement the Write command as per 6.3.2.10.3.3 and appendix C of standard.
		Cover encryption and CRC check and generation live in the G2_BinaryTag layer. '''

		passiveReplyTime=[CommandTime[1],CommandTime[1]+self.parseDelay]
		offReplyTime=[CommandTime[1],CommandTime[1]]
		replyTime=passiveReplyTime
		
		if self.state!='off':
			
			self.commandCount=self.commandCount+1		# increment command count since creation
			
			if self.lastCommand=='Req_RN':	#  proceed to parse the command, otherwise ignore it
				
				self.lastCommand='Write'
				if ((self.state=='arbitrate') or (self.state=='reply') or (self.state=='acknowledged')):	# go to arbitrate
					return ('NULL',passiveReplyTime)
					self.state='arbitrate'
				
				if ((self.state=='open') or (self.state=='secured')):
					#   write the data unless the info is locked or the location is outside the memory bank
					if (RN16==self.handle):	#  process the Write command
						try:	#  attempt the write, return an error if the relevant info is locked or the requested words are outside memory bounds
							locked=0		# initialize locked state
							if Membank==0:
								if (WordPtr<2):	#  this is a request to write the kill password
									if self.tagMemory.lockKill==1:
										locked=1
								if ((WordPtr>1) and (WordPtr<4)):   #  this request includes the access password
									if self.tagMemory.lockAccess==1:
										locked=1
							else:	#  check to see if the other banks are locked
								if self.tagMemory.lockBank[Membank]==1:	#  the bank requested is locked for writing
									locked=1
									print 'memory bank:', Membank, 'was locked'
									
							if locked==0:
								#  write the bytes to memory
								lowbyte=Data%256
								highbyte=int(Data/256)
								result1=self.tagMemory.writeMemoryByte(Membank,WordPtr,0,highbyte)
								result2=self.tagMemory.writeMemoryByte(Membank,WordPtr,1,lowbyte)
								replyTime[0]=CommandTime[1]+self.writeDelay
								replyTime[1]=replyTime[0]+self.replyDuration(33)
								return(0,RN16,'0000000000000000',replyTime)
		
							else:
								return (1,4,self.handle,'0000000000000000',passiveReplyTime)
										#  return error code for locked memory
						except ValueError:
							return (1,3,self.handle,'0000000000000000',passiveReplyTime)
								#  return error code for memory overrun
					else:
						# ignore the command if invalid handle, stay in same state
						return ('NULL',passiveReplyTime)	
				
				if self.state=='killed':
					return ('NULL',offReplyTime)
			
			else:		#  case of last command not Req_RN
				self.lastCommand='Write'
				if self.state!='killed':
					return('NULL',passiveReplyTime)
				else:
					return('NULL',offReplyTime)
				
		#case of self.state!='off'
		else:
			return  ['NULL',offReplyTime]
			
			
	#
	#	************************************************************
	#	Kill
	#	************************************************************
	#
	def Kill(self,Password,RFU,RN16,CRC16,CommandTime):
		#  do nothing yet

		passiveReplyTime=[CommandTime[1],CommandTime[1]+self.parseDelay]
		offReplyTime=[CommandTime[1],CommandTime[1]]
		replyTime=passiveReplyTime

		
		if self.state!='off':
			self.commandCount=self.commandCount+1		# increment command count since creation
			if ((self.state=='reply') or (self.state=='acknowledged')):
				if self.checkReplyTime(CommandTime)==False:		# if time since last reply > T2
					self.state='arbitrate'
		else:
			return  ['NULL',offReplyTime]
		
		
	#
	#	************************************************************
	#	Lock
	#	************************************************************
	#
	
	def Lock(self,Payload,RN16,CRC16,CommandTime):
		'''Implement the Lock command as per version 1.0.9 section 6.3.2.10.3.5 and appendix C. Note that Payload is presumed
		to be a list of bits rather than a number.'''


		passiveReplyTime=[CommandTime[1],CommandTime[1]+self.parseDelay]
		offReplyTime=[CommandTime[1],CommandTime[1]]
		replyTime=passiveReplyTime
		
		if self.state!='off':
			
			self.commandCount=self.commandCount+1		# increment command count since creation
			if (self.state=='ready'):
				return ('NULL',passiveReplyTime)
				
			if ((self.state=='reply') or (self.state=='acknowledged') or (self.state=='arbitrate')):
				self.state='arbitrate'
				return ('NULL',passiveReplyTime)
				
			if (self.state=='open'):
				return ('NULL',passiveReplyTime)
			
			if (self.state=='secured'):
				lockError=0
				if RN16==self.handle:	#   parse the command
					
					#  kill password lock
					
					if Payload[0]==1:	#   implement bit 10 of payload unless permalocked
						if self.tagMemory.permalockKill==1:	#  kill pwd is permalocked
							if Payload[10]==self.tagMemory.lockKill:   #  no change, that's ok
								pass	# do nothing
							else:	#  trying to change a lock bit of a permalocked pwd, send error
								lockError=4
	
						else:	#  kill password is not permalocked;  change settings as per mask
							self.tagMemory.lockKill=Payload[10]
					
					if Payload[1]==1:	#  implement bit 11 of payload unless permalocked
						if self.tagMemory.permalockKill==1:  	#  kill pwd is permalocked
							if Payload[11]==self.tagMemory.permalockKill:   # no change, that's ok
								pass	# do nothing
							else:
								lockError=4	
						else:	#  kill password is not permalocked;  change settings as per mask
							self.tagMemory.permalockKill=Payload[11]
								
					#  access password lock
					
					if Payload[2]==1:	#   implement bit 12 of payload unless permalocked
						if self.tagMemory.permalockAccess==1:	#  access pwd is permalocked
							if Payload[12]==self.tagMemory.lockAccess:   #  no change, that's ok
								pass	# do nothing
							else:	#  trying to change a lock bit of a permalocked pwd, send error
								lockError=4
	
						else:	#  access password is not permalocked;  change settings as per mask
							self.tagMemory.lockAccess=Payload[12]
					
					if Payload[3]==1:	#  implement bit 13 of payload unless permalocked
						if self.tagMemory.permalockAccess==1:  	#  access pwd is permalocked
							if Payload[13]==self.tagMemory.permalockAccess:   # no change, that's ok
								pass	# do nothing
							else:
								lockError=4	
						else:	#  access password is not permalocked;  change settings as per mask
							self.tagMemory.permalockAccess=Payload[13]		
							
					#   memory bank locks
					for j in range(1,4,1):
			
						if Payload[2+2*j]==1:	#   implement lock bit of payload unless permalocked
							if self.tagMemory.permalockBank[j]==1:	#  bank is permalocked
								if Payload[12+2*j]==self.tagMemory.lockBank[j]:   #  no change, that's ok
									pass	# do nothing
								else:	#  trying to change a lock bit of a permalocked bank, send error
									lockError=4
		
							else:	# bank is not permalocked;  change settings as per mask
								self.tagMemory.lockBank[j]=Payload[12+2*j]
						
						if Payload[3+2*j]==1:	#  implement lock bit of payload unless permalocked
							if self.tagMemory.permalockBank[j]==1:  	#  bank is permalocked
								if Payload[13+2*j]==self.tagMemory.permalockBank[j]:   # no change, that's ok
									pass	# do nothing
								else:   #  trying to change permalock bit of permalocked bank
									lockError=4	
							else:	#  bank is not permalocked;  change settings as per mask
								self.tagMemory.permalockBank[j]=Payload[13+2*j]		
							
					#  now return whatever happened
					if (lockError==0):	#  things went fine

						replyTime[0]=CommandTime[1]+self.lockDelay
						replyTime[1]=replyTime[0]+self.replyDuration(33)
						
						return (0,self.handle,'0000000000000000',replyTime)
					else:
						replyTime[0]=CommandTime[1]+self.T1
						replyTime[1]=replyTime[0]+self.replyDuration(41)
						return (1,lockError,self.handle,'0000000000000000',replyTime)
					
				else:				#  invalid handle, ignore the command
					return ('NULL',passiveReplyTime)
				
			if (self.state=='killed'):
				return ('NULL',offReplyTime)
				
		#case of self.state!='off':	
		else:
			return  ['NULL',offReplyTime]	
		
	#
	#	************************************************************
	#	Access
	#	************************************************************
	#
	def Access(self,CommandTime):
		#  do nothing yet
		
		if self.state!='off':
			
			self.commandCount=self.commandCount+1		# increment command count since creation
			if ((self.state=='reply') or (self.state=='acknowledged')):
				if self.checkReplyTime(CommandTime[1])==False:		# if time since last reply < T2
					self.state='arbitrate'
		
		else:
			return  ['NULL',[CommandTime[1],CommandTime[1]]]
		
	#
	#	************************************************************
	#	BlockWrite
	#	************************************************************
	#
	def BlockWrite(self,CommandTime):
		#  do nothing yet
		
		if self.state!='off':
			
			self.commandCount=self.commandCount+1		# increment command count since creation
			if ((self.state=='reply') or (self.state=='acknowledged')):
				if self.checkReplyTime(CommandTime[1])==False:		# if time since last reply < T2
					self.state='arbitrate'
					
		else:
			return  ['NULL',[CommandTime[1],CommandTime[1]]]
	
		
	#
	#	************************************************************
	#	BlockErase
	#	************************************************************
	#
	def BlockErase(self,CommandTime):
		#  do nothing yet
		if self.state!='off':
			self.commandCount=self.commandCount+1		# increment command count since creation
			if ((self.state=='reply') or (self.state=='acknowledged')):
				if self.checkReplyTime(CommandTime[1])==False:		# if time since last reply < T2
					self.state='arbitrate'
		else:
			return  ['NULL',[CommandTime[1],CommandTime[1]]]

		
	#
	#	************************************************************
	#	STATE FUNCTIONS
	#	************************************************************
	#
	
			
	#
	#	************************************************************
	#	PowerLost
	#	************************************************************
	#
	def PowerLost(self,CommandTime):
		'''This function takes actions associated with losing RF power:  sets timers for session and SL
		flags, set state back to off.'''
		self.state='off'
		#  set the flag time for the Select flag and session 0,2,3 flags (session 1 persistence
		#  is not dependent on power state and is not reset)
		
		self.SLFlagtime=CommandTime[1]
		for i in range(0,4,1):
			if i!=1:
				self.sessionFlagtimes[i]=CommandTime[1]
		

	#
	#	************************************************************
	#	PowerOn
	#	************************************************************
	#		
	def PowerOn(self,CommandTime):
		'''This function takes actions associated with RF power on:  checks timers for session and
		SL flags and sets them to their default states if appropriate, sets state to ready.'''
		
		if self.state=='off':	#  do nothing if the tag was already on
			self.state='ready'
			#   check flag state, initialize if needed
			#
			if (self.selectPersistence[1]!=0):
				if (CommandTime[1]-self.SLFlagtime>self.selectPersistence[1]):	# flag value lost, initialize
					self.SL=False
			#   reset the select time in any case on power up
			self.SLFlagtime=CommandTime[1]
			
			#  now do the same exercise for the session flags
			
	
			# !!!!!!!!!!!!!!!!!!!!!!! DEBUG
			#
			#print('at tag on, command time is:', CommandTime,' and session flag times are:',self.sessionFlagtimes)
			
			
			if (self.session0Persistence[1]!=0):
				if(CommandTime[1]-self.sessionFlagtimes[0])>self.session0Persistence[1]:   # flag lost, initialize
					self.sessionFlags[0]='A'
			self.sessionFlagtimes[0]=CommandTime[1]
			
			if (self.session1Persistence[1]!=0):
				if (CommandTime[1]-self.sessionFlagtimes[1])>self.session1Persistence[1]:   # flag lost, initialize
					self.sessionFlags[1]='A'
			self.sessionFlagtimes[1]=CommandTime[1]
			
			if (self.session2Persistence[1]!=0):
				if (CommandTime[1]-self.sessionFlagtimes[2])>self.session1Persistence[1]:   # flag lost, initialize
					self.sessionFlags[2]='A'
			self.sessionFlagtimes[2]=CommandTime[1]
			
			if (self.session3Persistence[1]!=0):
				if (CommandTime[1]-self.sessionFlagtimes[3])>self.session1Persistence[1]:   # flag lost, initialize
					self.sessionFlags[3]='A'
			self.sessionFlagtimes[3]=CommandTime[1]
			
			#   calculate CRC and write it to the first word of the EPC bank
			returnlist=self.getPC_EPC()
			PCEPCbytes=len(returnlist)
			CRCvalue=CRCs.CRC16(returnlist,PCEPCbytes,0xFFFF)
			CRChighbyte=int(CRCvalue/256)
			CRClowbyte=CRCvalue%256
			result1=self.tagMemory.writeMemoryByte(1,0,0,CRChighbyte)
			result2=self.tagMemory.writeMemoryByte(1,0,1,CRClowbyte)

			
			#  initialize those variables that were lost at power off
			self.SLtarget=0	       #  default to any SL flag is ok
			self.lastSession='NULL'     # session is not meaningful until we receive a Query
			self.currentQ=0
			self.sessionTarget='A'		#  default target to A
			self.powerFlag=1
			self.lastCommand='NULL'	#  needed for Write
			self.RTcal= 75		#  all time intervals in microseconds
			self.TRcal= 80
			self.Tari=25
			self.DivideRatio= 8
			self.Mvalue = 1	#  MMS multiplier
			self.Tpri=self.TRcal/self.DivideRatio	#  standard 1.0.9  table 6.3.1.3.3
			if self.RTcal>10*self.Tpri:
				self.T1=self.RTcal
			else:
				self.T1=10*self.Tpri
			self.T2max=20*self.Tpri
			self.TRext = 0
			self.LinkFreq = 1000/self.Tpri	# in KHz
			self.Truncate= 0    #  from Select command
			self.TruncatePointer=0
			#      16-bit random numbers
			self.QRN=0x00	#  this is the random number returned when the slot counter comes up 0
			self.handle=0x00	#  this is the random number returned by a Req_RN
			self.cover=0x00    	#  this is the current RN for cover coding   
			self.slotCounter=0
			self.halfDead=0	#  this is used to note when the first of two
						#  KILL commands was successful 
			self.halfAccessed = 0	#  this is used to note when the first of two
						#  Access commands was successful
			
				
			return('NULL',CommandTime[1]+self.parseDelay)

		#  end case of self.state=='off'
	#  end PowerOn
		
#	end class G2_LogicalTag


#  G2_LogicalTag  
#  version 2.00.080415.1449
#  Daniel M. Dobkin   
#  Enigmatics
#  www.enigmatic-consulting.com
		
